/*
 * EJS Embedded JavaScript templates
 * Copyright 2112 Matthew Eernisse (mde@fleegix.org)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/

var utils = require('./utils')
  , ejs
  , templateCache = {}
  , jsCache = {}
  , includeFile
  , includeSource
  , resolveInclude
  , compile
  , rethrow
  , _DEFAULT_DELIMITER = '%'
  , _REGEX_STRING = '(<%%)|(<%=)|(<%-)|(<%#)|(<%)|(%>)|(-%>)'
  , _OPTS_MAP = {
      cache: 'cache'
    , filename: 'filename'
    , delimiter: 'delimiter'
    , scope: 'context'
    , context: 'context'
    , debug: 'debug'
    , compileDebug: 'compileDebug'
    , client: 'client'
    };

resolveInclude = function (name, filename) {
  var path = require('path')
    , dirname = path.dirname
    , extname = path.extname
    , join = path.join
    , includePath = join(dirname(filename), name)
    , ext = extname(name);
  if (!ext) {
    includePath += '.ejs';
  }
  return includePath;
};

includeFile = function (path, options) {
  var fn
    , opts = utils.shallowCopy({}, options || {})
    , fs = require('fs')
    , includePath
    , template;
  if (!opts.filename) {
    throw new Error('`include` requires the \'filename\' option.');
  }
  includePath = resolveInclude(path, opts.filename);
  if (opts.cache) {
    template = templateCache[includePath];
    if (!template) {
      template = fs.readFileSync(includePath).toString().trim();
      templateCache[includePath] = template;
    }
  }
  else {
    template = fs.readFileSync(includePath).toString().trim();
  }

  opts.filename = includePath
  fn = exports.compile(template, opts);
  return fn;
};

includeSource = function (path, options) {
  var fn
    , opts = utils.shallowCopy({}, options || {})
    , fs = require('fs')
    , includePath
    , template;
  if (!opts.filename) {
    throw new Error('`include` requires the \'filename\' option.');
  }
  includePath = resolveInclude(path, opts.filename);
  if (opts.cache) {
    template = templateCache[includePath];
    if (!template) {
      template = fs.readFileSync(includePath).toString().trim();
      templateCache[includePath] = template;
    }
  }
  else {
    template = fs.readFileSync(includePath).toString().trim();
  }

  opts.filename = includePath
  var templ = new Template(template, opts || {});
  templ.generateSource();
  return templ.source;
};

function rethrow(err, str, filename, lineno){
  var lines = str.split('\n')
    , start = Math.max(lineno - 3, 0)
    , end = Math.min(lines.length, lineno + 3);

  // Error context
  var context = lines.slice(start, end).map(function(line, i){
    var curr = i + start + 1;
    return (curr == lineno ? ' >> ' : '    ')
      + curr
      + '| '
      + line;
  }).join('\n');

  // Alter exception message
  err.path = filename;
  err.message = (filename || 'ejs') + ':'
    + lineno + '\n'
    + context + '\n\n'
    + err.message;

  throw err;
}

compile = exports.compile = function (template, opts) {
  var templ = new Template(template, opts || {});
  return templ.compile();
};

// template, [data], [opts]
// Have to include an empty data object if you want opts and no data
exports.render = function () {
  var args = Array.prototype.slice.call(arguments)
    , template = args.shift()
    , data = args.shift() || {}
    , opts = args.shift() || {}
    , fn;

  // No options object -- if there are optiony names
  // in the data, copy them to options
  if (arguments.length == 2) {
    for (var p in _OPTS_MAP) {
      if (typeof data[p] != 'undefined') {
        opts[_OPTS_MAP[p]] = data[p];
      }
    }
  }
  // v1 compat
  // 'scope' is 'context'
  // FIXME: Remove this in a future version
  if (opts.scope) {
    opts.context = opts.scope;
    delete opts.context;
  }

  if (opts.cache) {
    if (!opts.filename) {
      throw new Error('cache option requrires a filename');
    }
    fn = jsCache[filename];
    if (!fn) {
      fn = exports.compile(template, opts);
      jsCache[filename] = fn;
    }
  }
  else {
    fn = exports.compile(template, opts);
  }
  return fn.call(opts.context, data);
};

// path, [data] [opts], cb
// Have to include an empty data object if you want opts and no data
exports.renderFile = function () {
  var read = require('fs').readFile
    , args = Array.prototype.slice.call(arguments)
    , path = args.shift()
    , cb = args.pop()
    , data = args.shift() || {}
    , opts = args.pop() || {}
    , template
    , handleTemplate;

  handleTemplate = function (template) {
    var result;
    try {
      result = exports.render(template, data, opts);
    }
    catch(err) {
      cb(err);
    }
    cb(null, result);
  };

  template = templateCache[path];
  if (opts.cache && template) {
    handleTemplate(template);
  }
  else {
    read(path, function (err, data) {
      var tmpl = data.toString().trim();
      if (err) {
        return cb(err);
      }
      if (opts.cache) {
        templateCache[path] = tmpl;
      }
      handleTemplate(tmpl);
    });
  }
};

exports.clearCache = function () {
  templateCache = {};
  jsCache = {};
};

var Template = function (text, opts) {
  var options = {};
  this.templateText = text;
  this.mode = null;
  this.truncate = false;
  this.currentLine = 1;
  this.source = '';
  options.client = opts.client || false;
  options.escapeFunction = opts.escape || utils.escapeXML;
  options.compileDebug = opts.compileDebug !== false;
  options.filename = opts.filename;
  options.delimiter = opts.delimiter || exports.delimiter || _DEFAULT_DELIMITER;
  this.opts = options;

  this.regex = this.createRegex();
};

Template.prototype = new function () {

  this.modes = {
      EVAL: 'eval'
    , ESCAPED: 'escaped'
    , RAW: 'raw'
    , APPEND: 'append'
    , COMMENT: 'comment'
    , LITERAL: 'literal'
  };

  this.createRegex = function () {
    var str = _REGEX_STRING
      , delim = utils.escapeRegExpChars(this.opts.delimiter);
    str = str.replace(/%/g, delim);
    return new RegExp(str);
  };

  this.compile = function () {
    var self = this
      , src
      , fn
      , opts = this.opts
      , escape = opts.escapeFunction;

    if (!this.source) {
      this.generateSource();
      this.source = 'var __output = ""; with (locals) { ' + this.source;
      this.source += '} return __output.trim();';
    }

    if (opts.compileDebug) {
      src = 'var __line = 1' +
          ', __lines = ' + JSON.stringify(this.templateText) +
          ', __filename = ' + (opts.filename ?
                JSON.stringify(opts.filename) : 'undefined') +
          '; try {' +
          this.source + '} catch (e) { rethrow(e, __lines, __filename, __line); }';
    }
    else {
      src = this.source;
    }

    if (opts.debug) {
      console.log(src);
    }

    if (opts.client) {
      src = 'escape = escape || ' + escape.toString() + ';\n' + src;
      src = 'rethrow = rethrow || ' + rethrow.toString() + ';\n' + src;
    }

    try {
      fn = new Function('locals, escape, include, rethrow', src);
    }
    catch(e) {
      if (e instanceof SyntaxError) {
        if (opts.filename) {
          e.message += ' in ' + opts.filename;
        }
        e.message += ' while compiling ejs';
        throw e;
      }
    }

    if (opts.client) {
      return fn;
    }

    // Return a callable function which will execute the function
    // created by the source-code, with the passed data as locals
    return function (data) {
      var include = function (path, includeData) {
        var d = utils.shallowCopy({}, data);
        if (includeData) {
          d = utils.shallowCopy(d, includeData);
        }
        return includeFile(path, opts).apply(this, [d]);
      };
      return fn.apply(this, [data || {}, escape, include, rethrow]);
    };

  };

  this.generateSource = function () {
    var self = this
      , matches = this.parseTemplateText()
      , d = this.opts.delimiter;

    if (matches && matches.length) {
      matches.forEach(function (line, index) {
        var closing
          , include
          , includeOpts
          , includeSrc;
        // If this is an opening tag, check for closing tags
        // FIXME: May end up with some false positives here
        // Better to store modes as k/v with '<' + delimiter as key
        // Then this can simply check against the map
        if (line.indexOf('<' + d) === 0) {
          closing = matches[index + 2];
          if (!(closing == d + '>' || closing == '-' + d + '>')) {
            throw new Error('Could not find matching close tag for "' + line + '".');
          }
        }
        // HACK: backward-compat `include` preprocessor directives
        if ((include = line.match(/^\s*include\s+(\S+)/))) {
          includeOpts = utils.shallowCopy({}, self.opts);
          includeSrc = includeSource(include[1], includeOpts);
          includeSrc = ";(function(){" + includeSrc + "})();";
          self.source += includeSrc;
        }
        else {
          self.scanLine(line);
        }
      });
    }

  };

  this.parseTemplateText = function () {
    var str = this.templateText
      , pat = this.regex
      , result = pat.exec(str)
      , arr = []
      , firstPos
      , lastPos;

    while (result) {
      firstPos = result.index;
      lastPos = pat.lastIndex;

      if (firstPos !== 0) {
        arr.push(str.substring(0, firstPos));
        str = str.slice(firstPos);
      }

      arr.push(result[0]);
      str = str.slice(result[0].length);
      result = pat.exec(str);
    }

    if (str !== '') {
      arr.push(str);
    }

    return arr;
  };

  this.scanLine = function (line) {
    var self = this
      , d = this.opts.delimiter
      , newLineCount = 0
      , _addOutput;

    _addOutput = function () {
      if (self.truncate) {
        line = line.replace('\n', '');
      }

      // Preserve literal slashes
      line = line.replace(/\\/g, '\\\\');

      // Convert linebreaks
      line = line.replace(/\n/g, '\\n');
      line = line.replace(/\r/g, '\\r');

      // Escape double-quotes
      // - this will be the delimiter during execution
      line = line.replace(/"/g, '\\"');
      self.source += ';__output += "' + line + '";';
    };

    newLineCount = (line.split('\n').length - 1);

    switch (line) {
      case '<' + d:
        this.mode = this.modes.EVAL;
        break;
      case '<' + d + '=':
        this.mode = this.modes.ESCAPED;
        break;
      case '<' + d + '-':
        this.mode = this.modes.RAW;
        break;
      case '<' + d + '#':
        this.mode = this.modes.COMMENT;
        break;
      case '<' + d + d:
        this.mode = this.modes.LITERAL;
        this.source += ';__output += "' + line.replace('<' + d + d, '<' + d) + '";';
        break;
      case d + '>':
      case '-' + d + '>':
        if (this.mode == this.modes.LITERAL) {
          _addOutput();
        }

        this.mode = null;
        this.truncate = line.indexOf('-') == 0;
        break;
      default:
        // In script mode, depends on type of tag
        if (this.mode) {
          // Strip double-slash comments in all exec modes
          switch (this.mode) {
            case this.modes.EVAL:
            case this.modes.ESCAPED:
            case this.modes.RAW:
              line = line.replace(/\/\/.*$/, '');
          }
          switch (this.mode) {
            // Just executing code
            case this.modes.EVAL:
              this.source += line;
              break;
            // Exec, esc, and output
            case this.modes.ESCAPED:
              // Add the exec'd, escaped result to the output
              this.source += ';__output += escape(' +
                  line.replace(/;\S*/, '').trim() + ');';
              break;
            // Exec and output
            case this.modes.RAW:
              // Add the exec'd result to the output
              this.source += ';__output += ' + line.trim() + ';';
              break;
            case this.modes.COMMENT:
              // Do nothing
              break;
            // Literal <%% mode, append as raw output
            case this.modes.LITERAL:
              _addOutput();
              break;
          }
        }
        // In string mode, just add the output
        else _addOutput();
    }

    if (newLineCount) {
      this.currentLine += newLineCount;
      this.source += ';__line = ' + this.currentLine + ';';
    }
  };
};

// Express support
exports.__express = exports.renderFile;

// Add require support
if (require.extensions) {
  require.extensions['.ejs'] = function (module, filename) {
    filename = filename || module.filename;
    var fs = require('fs')
      , options = {
          filename: filename
        , client: true
        }
      , template = fs.readFileSync(filename).toString().trim()
      , fn = compile(template, options);
    module._compile('module.exports = ' + fn.toString() + ';', filename);
  };
}

if (typeof window != 'undefined') {
  window.ejs = exports;
}
